﻿/**
 * Authors: http://d.hatena.ne.jp/ultraist/
 */
module common.downloadlist;

// ダウンロードリスト

private import std.socket;
private import std.c.time;
private import std.path;
private import pes.system.all;
private import pes.conv;
private import worker.cachebitmap;

// アドレス
class Distributor
{
public:
	enum State:uint 
	{
		DEAD,		// ノード死亡(禁止)
		ACTIVE,		// イケてる(即DL)
		BUSY,		// 忙しい(数分ウェイト)
	}
	
protected:
	InternetAddress		m_address;
	uint				m_wait_time;
	int					m_priority	= 0x80;
	State				m_state		= State.ACTIVE;
	int					m_lock;
	bool				m_virgin;
	
public:
	this(InternetAddress addr)
	{
		m_address 	= addr;
		m_virgin	= true;
	}
	
	bool isLock()
	{
		return m_lock != 0;
	}
	
	void lock()
	{
		++m_lock;
		m_virgin = false;
	}
	
	void unlock()
	{
		--m_lock;
		if (m_lock < 0) {
			assert(0);
		}
	}
	
	InternetAddress address()
	{
		return m_address;
	}
	
	State state()
	{
		return m_state;
	}
	void active()
	{
		m_state = State.ACTIVE;
	}
	
	void die()
	{
		m_state = State.DEAD;
	}
	
	void busy()
	{
		m_state = State.BUSY;
	}
	
	bool isActive()
	{
		if (m_state == State.BUSY) {
			if (m_wait_time < time(null)) {
				active();
			}
		}
		
		return m_state == State.ACTIVE;
	}
	
	bool isDead()
	{
		return m_state == State.DEAD;
	}
	
	bool isBusy()
	{
		if (m_state == State.BUSY) {
			if (m_wait_time < time(null)) {
				active();
			}
		}
		
		return m_state == State.BUSY;
	}
	
	int priority()
	{
		if (m_virgin) {
			return 0x100;
		}
		
		return m_priority;
	}
	
	void priority(int pri)
	{
		m_virgin = false;
		
		if (pri > 0xff) {
			pri = 0xff;
		} else if (pri < 0) {
			pri = 0;
		}
		
		m_priority = pri;
	}
	
	uint waitTime()
	{
		int wt = m_wait_time - time(null);
		
		if (wt < 0) {
			active();
			return 0;
		}
		
		return wt;
	}
	
	void waitTime(uint wt)
	{
		m_wait_time = wt + time(null);
	}
	
	override int opCmp(Object o)
	{
		Distributor rhs = cast(Distributor)o;
		
		return rhs.priority - priority;
	}
}

// アイテム
class DownloadItem
{
protected:
	ubyte[]				m_hash;
	ulong				m_size;
	char[]				m_name;
	Distributor[ulong]	m_address_list;
	uint				m_priority  = 0xff;
	int					m_downloading;
	uint				m_block_max	= 64; // 64blocks
	bool				m_completed;
	bool				m_auto_download;
	time_t				m_ttl;		// auto downloadを消すUNIX時間
	CacheBitmap			m_bitmap;
	Object				m_cache_lock;
	
public:
	ulong toKey(InternetAddress address)
	{
		ulong key = (cast(ulong)address.addr) << 32;
		key |= address.port;
		
		return key;
	}
	
	this(ubyte[] hash, ulong sz, char[] name)
	{
		m_hash 	= hash;
		m_size	= sz;
		m_name	= getBaseName(name);
		
		m_cache_lock = new Object;
	}
	
	void addAddress(InternetAddress address)
	{
		ulong key = toKey(address);
		
		if ((key in m_address_list) == null) {
			m_address_list[key] = new Distributor(address);
		}
	}
	
	void addAddress(Distributor address)
	{
		ulong key = toKey(address.address);
		
		if ((key in m_address_list) == null) {
			m_address_list[key] = address;
		}
	}
	
	void clearAddress()
	{
		foreach (ulong key; m_address_list.keys) {
			m_address_list.remove(key);
		}
	}
	
	char[] name()
	{
		if (m_name.length == 0) {
			return binaryToHexString(m_hash);
		}
		
		return m_name;
	}
	
	void name(char[] n)
	{
		m_name = getBaseName(n);
	}
	
	Distributor[] address()
	{
		return m_address_list.values;
	}
	
	bool addressExists(InternetAddress addr)
	{
		return ((toKey(addr) in m_address_list) != null);
	}
	
	Distributor distributor(InternetAddress addr)
	{
		return m_address_list[toKey(addr)];
	}
	
	ubyte[] hash()
	{
		return m_hash;
	}
	
	ulong dataSize()
	{
		return m_size;
	}
	
	ulong blocks()
	{
		return (m_size / 0x10000) + ((m_size % 0x10000) ? 1:0);
	}
	
	uint distributorCount()
	{
		return m_address_list.length;
	}
	
	ulong blockCount()
	{
		ulong count;
		synchronized(this) {
			count = m_bitmap.count(blocks);
		}
		
		return count;
	}
	
	
	uint downloadBlockMax()
	{
		return m_block_max;
	}
	
	void downloadBlockMax(uint max)
	{
		if (max < 513) {
			m_block_max = max;
		} else {
			m_block_max = 512;
		}
	}
	
	void downloadBlockMaxUp()
	{
		downloadBlockMax = downloadBlockMax * 2; // 64 128 256 512
	}
	
	void downloadBlockMaxReset()
	{
		m_block_max = m_block_max.init;
	}
	
	void tryDownload()
	{
		if (m_priority == 0) {
			m_priority = 0xff;
		} else {
			--m_priority;
		}
		++m_downloading;
	}
	
	bool isDownloading()
	{
		return m_downloading > 0;
	}
	
	void endDownload()
	{
		--m_downloading;
		if (m_downloading < 0) {
			assert(0);
		}
	}
	
	void complete()
	{
		m_completed = true;
		m_priority = 0;
	}
	
	bool isComplete()
	{
		return m_completed;
	}
	
	override int opCmp(Object o)
	{
		DownloadItem rhs = cast(DownloadItem)o;
		
		return (cast(long)rhs.m_priority) - (cast(long)m_priority);
	}
	
	void autoDownloadItem(bool b)
	{
		m_auto_download = b;
	}
	
	bool isAutoDownloadItem()
	{
		return m_auto_download;
	}
	
	void ttl(time_t ttl)
	{
		m_ttl = ttl;
	}
	
	time_t ttl()
	{
		return m_ttl;
	}
	
	void bitSet(uint bitn) {
		synchronized(this) {
			m_bitmap.set(bitn);
		}
	}
	
	void mergeBitmap(CacheBitmap bitmap)
	{
		synchronized(this) {
			m_bitmap.merge(bitmap);
		}
	}
	
	Object cacheLock()
	{
		return m_cache_lock;
	}
	
	void setBit(ulong i)
	{
		m_bitmap.set(i);
	}
}

// リスト
class DownloadList: Locker!(ubyte[], DownloadItem)
{
protected:
	DownloadItem[ubyte[]]	m_download_item;
	
public:
	this(SemaphoreResource resource, semaphore id)
	{
		super(resource, id);
	}
	
	override void set(ubyte[] hash, DownloadItem item)
	{
		signal();
		try {
			if (!exists(hash)) {
				super.set(hash, item);
			}
		} finally {
			release();
		}
	}
}






