﻿/**
 * Authors: http://d.hatena.ne.jp/ultraist/
 */
module worker.cachebitmap;


// キャッシュビットマップ
// 

private import pes.byteorder;

struct CacheBitmap
{
align(1):
	uint	map[4096];
	
	void set(ulong no)
	{
		ulong index 	= no / (uint.sizeof * 8);
		ulong bits		= no % (uint.sizeof * 8);
		
		if (index >= map.length) {
			throw new Exception("Cache bit map overflow");
		}
		
		map[index] = hostToNetwork32(networkToHost32(map[index]) | (0x80000000 >> bits));
	}
	
	void unset(ulong no)
	{
		ulong index 	= no / (uint.sizeof * 8);
		ulong bits		= no % (uint.sizeof * 8);
		
		if (index >= map.length) {
			throw new Exception("Cache bit map overflow");
		}
		
		map[index] = hostToNetwork32(networkToHost32(map[index]) & ~(0x80000000 >> bits));
	}
	
	bool isSet(ulong no)
	{
		ulong index 	= no / (uint.sizeof * 8);
		ulong bits		= no % (uint.sizeof * 8);
		
		if (index >= map.length) {
			throw new Exception("Cache bit map overflow");
		}
		
		return (networkToHost32(map[index]) & (0x80000000 >> bits)) != 0;
	}
	
	// 埋まっていない断片を[2]な配列で返す
	// [0] = 始まり, [1] = 終了
	// ただし[0]は含むが、[1]は含まない(未満)
	ulong[][] fragments(ulong data_size)
	{
		ulong max = (data_size / 0x10000) + ((data_size % 0x10000 != 0) ? 1:0);
		bool  scan = false;
		ulong begin;
		ulong[][] 	frags;
		ulong[2]	pair;
		
		for (ulong i = 0; i < max ; ++i) {
			if (scan) {
				if (this.isSet(i)) {
					pair[0] = begin;
					pair[1] = i;
					frags ~= pair.dup;
					scan	= false;
				}
			} else {
				if (!this.isSet(i)) {
					scan = true;
					begin = i;
				}
			}
		}
		if (scan) {
			pair[0] = begin;
			pair[1] = max;
			frags ~= pair.dup;
		}
		
		return frags;
	}
	
	void clear()
	{
		for (uint i = 0; i < map.length; ++i) {
			map[i] = 0;
		}
	}
	
	void fill(ulong data_size)
	{
		ulong 	blockno 	= (data_size / 0x10000) + ((data_size % 0x10000 != 0) ? 1:0);
		ulong	endindex	= blockno / (uint.sizeof * 8);
		ulong	remainder	= blockno % (uint.sizeof * 8);
		
		for (ulong i = 0; i < endindex; ++i) {
			map[i] = ~0;
		}
		if (remainder != 0) {
			for (ulong i = 0; i < remainder; ++i) {
				set((endindex * uint.sizeof * 8) + i);
			}
		}
	}
	
	ulong count(ulong block_max)
	{
		ulong cnt;
		for (ulong i = 0; i < block_max; ++i) {
			if (isSet(i)) {
				++cnt;
			}
		}
		
		return cnt;
	}
	
	ubyte[] toBlob()
	{
		return (cast(ubyte*)this)[0 .. (*this).sizeof];
	}
	
	void merge(CacheBitmap bitmap)
	{
		for (uint i = 0; i < (map.length * (uint.sizeof * 8)); ++i) {
			if (bitmap.isSet(i)) {
				set(i);
			}
		}
	}
}

debug(cachebitmap)
{
	import std.stdio;
	import pes.dump;
	
	int main()
	{
		CacheBitmap bitmap, bitmap2;
		ulong[][] frags;
		
		bitmap.clear();
		bitmap2.clear();
		bitmap.set(3);
		bitmap.set(34);
		bitmap.set(197);
		bitmap.set(1024);
		bitmap.set(2048);
		bitmap.set(2049);
		bitmap.set(4095);
		bitmap.set(4096);
		bitmap.set(4097);
		bitmap.set(4098);
		bitmap2.merge(bitmap);
		
		for (int i = 0; i < 5000; ++i) {
			if (bitmap.isSet(i)) {
				writef("%s: 1\n", i);
			}
		}
		writef("\n");

		frags = bitmap.fragments(0x100000);
		foreach (ulong[] e; frags) {
			writef("%x %x\n", e[0], e[1]);
		}
		
		
		return 0;
	}
}

