﻿/**
 * Authors: http://d.hatena.ne.jp/ultraist/
 */
module worker.cacheencoder;

// キャッシュエンコーダー

private import pes.data.sqlite3;
private import std.md5;
private import std.file;
private import std.path;
private import std.thread;
private import std.base64;
private import std.c.time;
private import std.mmfile;
private import std.stream;
private import std.random;
private import std.gc;

private import pes.checksum;
private import pes.rc4;
private import pes.conv;
private import pes.data.sqlite3;
private import pes.system.all;

private import worker.cachebitmap;
private import common.workertable;
private import common.environment;

private const char[] CACHE_BLOCK_TABLE_CREATE = 
	"CREATE TABLE cache_block_table(no int64, hash blob, block blob)";

private const char[] CACHE_BLOCK_INDEX_CREATE =
	"CREATE UNIQUE INDEX cache_block_index on cache_block_table(no)";

private const char[] CACHE_BLOCK_INSERT =
	"INSERT INTO cache_block_table(no, hash, block) values(?, ?, ?)";

class CacheEncoder: WorkerTask
{
protected:
	char[] 			m_filepath;
	Environment 	s_environ;
	ubyte[]			m_hash;
	
	char[]			m_errmsg;
	int				m_result;
	
	const ubyte[] CACHE_KEY_INIT = [0x68, 0x65, 0x61, 0x64, 0x65, 0x72];
	
	
	int mainLoop()
	{
		setPriority(PRIORITY.DECREASE);
		gcLimit = 16;
		
		try {
			m_result = threadMain();
		} catch (Exception e) {
			m_errmsg = e.toString;
			m_result = -1;
		}
		
		std.gc.fullCollect();
		
		return m_result;
	}
	
	int threadMain()
	{
		char[] filepath = filePath(m_filepath);
		
		name = std.path.getBaseName(m_filepath);
		
		if (std.file.exists(filepath)) {
			return 0; // 既に存在するため終了
		}
		encode(m_filepath, filepath, hash);
		
		return 0;
	}
	
	void toValidPath(char[] path)
	{
		for (int i = 0; i < path.length; ++i) {
			if (path[i] == '=') {
				path[i] = '_';
			} else if (path[i] == '/') {
				path[i] = '-';
			} else if (path[i] == '+') {
				path[i] = ' ';
			}
		}
	}
	
	char[] filePath(char[] filepath)
	{
		m_hash				= md5FileSum(filepath);
		uint	blen		= std.base64.encodeLength(m_hash.length);
		char[]	name		= new char[blen];
		
		name = std.base64.encode(cast(char[])m_hash, name);
		toValidPath(name);
		
		return s_environ.configuration.cacheFolder ~ std.path.sep ~ name ~ ".db";
	}
	
	
	void encode(char[] file, char[] cache, ubyte[] hash)
	{
		RC4				rc4ctx		= new RC4;
//		MmFile 			mmap	 	= new MmFile(file);
//		MmFileStream	stream		= new MmFileStream(mmap);
		ubyte[]			key			= CACHE_KEY_INIT;
		char[]			text_hash	= binaryToHexString(hash);
		File			stream		= new File(file);
		SQLite			db			= new SQLite(cache);
		SQLiteStatement stmt;
		ubyte[0x10000]	buffer;
		ubyte[16]		sum;
		uint			len;
		uint			block_count	= 0;
		ulong			data_size;
		
		// 仕事量設定
		workMax(cast(uint)std.file.getSize(file));
		
		// キャッシュファイル作成
		try {
			db.begin();
			
			db.exec(CACHE_BLOCK_TABLE_CREATE);
			db.exec(CACHE_BLOCK_INDEX_CREATE);
			
			stmt = db.prepare(CACHE_BLOCK_INSERT);
			
			data_size = 0;
			while ((len = stream.readBlock(buffer.ptr, buffer.length)) != 0) {
				data_size += len;
				
				rc4ctx.key(key);
				std.md5.sum(sum, buffer[0 .. len]);
				rc4ctx.crypt(buffer[0 .. len]);
				
				stmt.bind(1, block_count, sum, buffer[0 .. len]);
				
				stmt.step();
				stmt.reset();
				
				++block_count;
				key = cast(ubyte[])std.string.format("%d %s", block_count, text_hash);
				
				work(len); // 進み具合更新
				
				pollTerminateEvent(); // 中断チェック
				
				super.fullCollect();
				delayTask(10);
			}
		} finally {
			stream.close();
			stmt.finalize();
			db.commit();
			db.close();
		}
		
		// インデックスに登録
		insertHeader(file, hash, data_size);
	}
	
	void insertHeader(char[] file, ubyte[] hash, ulong data_size)
	{
		CacheHeader header = new CacheHeader;
		
		header.bitmap.fill(data_size);
		header.hash 			= hash;
		header.dataSize			= data_size;
		header.fileName			= std.path.getBaseName(file);
		header.modifiedTime		= cast(uint)(time(null) - (32400 * 2));
//		header.referredCount	= 10 + (rand() % 10);
		header.completed		= 1;
		
		// 既存の場合は消す
		s_environ.cacheHeader.remove(hash, data_size);
		s_environ.cacheHeader.set(header);
	}
	
public:
	this(TaskResource resource, char[] filepath, Environment environ)
	{
		super(ANONYMOUS_MAIL, resource);
		m_filepath 	= filepath;
		s_environ	= environ;
	}
	
	int exitState()
	{
		return m_result;
	}
	
	char[] errorMessage()
	{
		return m_errmsg;
	}
	
	ubyte[] hash()
	{
		return m_hash;
	}
}
/*
debug(cache_task) {
	import std.stdio;
	import std.c.time;
int main()
{
	Configuration config = new Configuration;
	CacheEncoder encoder = new CacheEncoder("up\\test.txt", config);
	SQLite db = new SQLite(config.cacheIndex);
	time_t start = time(null);
	
	try {
		db.exec(CACHE_HEADER_TABLE_CREATE);
	} catch {
		
	}
	
	encoder.start();
	encoder.wait();
	
	writef("-- end %s sec\n", time(null) - start);
	
	return 0;
	
}
}
*/